"""
The module implements exceptions for Zserio python runtime library.
"""

class PythonRuntimeException(Exception):
    """
    Exception thrown in case of an error in Zserio python runtime library.
    """
